From ba9fd7d02bc0a11daa6a4a8b79b9c4d6b84e1fc4 Mon Sep 17 00:00:00 2001
From: Sebastien Zwickert <dilaroga@free.fr>
Date: Tue, 11 May 2010 13:23:13 +0200
Subject: [PATCH] VDADecoder  module.

---
 configure.ac             |   12 +
 modules/codec/Modules.am |    1 +
 modules/codec/vda.c      |  547 ++++++++++++++++++++++++++++++++++++++++++++++
 3 files changed, 560 insertions(+), 0 deletions(-)
 create mode 100755 modules/codec/vda.c

diff --git a/configure.ac b/configure.ac
index 7c75b82..f8d06d7 100644
--- a/configure.ac
+++ b/configure.ac
@@ -3092,6 +3092,18 @@ AC_CHECK_HEADERS(png.h, [
 fi
 AM_CONDITIONAL(BUILD_OSDMENU, [test "${enable_png}" != "no"])
 
+ dnl
+dnl vda decoder module
+dnl
+AC_ARG_ENABLE(vda,
+  [  --enable-vda            VDA support (default auto)])
+if test "${enable_vda}" != "no"; then
+AC_CHECK_HEADERS(VideoDecodeAcceleration/VDADecoder.h, [
+    VLC_ADD_PLUGIN([vda]),
+    VLC_ADD_LDFLAGS([vda],[-Wl,-framework,CoreFoundation,-framework,VideoDecodeAcceleration,-framework,QuartzCore,$SWSCALE_LIBS])
+  ])
+fi
+
 dnl
 dnl H264 encoder plugin (using libx264)
 dnl
diff --git a/modules/codec/Modules.am b/modules/codec/Modules.am
index 44ddcba..7d56af9 100644
--- a/modules/codec/Modules.am
+++ b/modules/codec/Modules.am
@@ -21,6 +21,7 @@ SOURCES_x264 = x264.c
 SOURCES_twolame = twolame.c
 SOURCES_dirac = dirac.c
 SOURCES_png = png.c
+SOURCES_vda = vda.c
 SOURCES_svcdsub = svcdsub.c
 SOURCES_cvdsub = cvdsub.c
 SOURCES_fake = fake.c
diff --git a/modules/codec/vda.c b/modules/codec/vda.c
new file mode 100755
index 0000000..cf8c335
--- /dev/null
+++ b/modules/codec/vda.c
@@ -0,0 +1,547 @@
+/*****************************************************************************
+ * vda.c: Video Decoder Acceleration.
+ *****************************************************************************
+ * Copyright (C) 2010 Sebastien Zwickert
+ * $Id$
+ *
+ * Authors: Sebastien Zwickert <dilaroga (at) free fr>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+/*****************************************************************************
+ * Preamble
+ *****************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <vlc_common.h>
+#include <vlc_plugin.h>
+#include <vlc_codec.h>
+#include <vlc_cpu.h>
+
+#ifdef HAVE_LIBSWSCALE_SWSCALE_H
+#   include <libswscale/swscale.h>
+#   include <libavcodec/avcodec.h>
+#elif defined(HAVE_FFMPEG_SWSCALE_H)
+#   include <ffmpeg/swscale.h>
+#   include <ffmpeg/avcodec.h>
+#endif
+
+#include <VideoDecodeAcceleration/VDADecoder.h>
+
+/*****************************************************************************
+ * vlc_va_frame_t : vda frame descriptor
+ *****************************************************************************/
+typedef struct
+{
+    int64_t                 i_pts;
+    CVPixelBufferRef        buffer;
+    struct vlc_va_frame_t   *p_next ;
+
+} vlc_va_frame_t;
+
+/*****************************************************************************
+ * decoder_sys_t : vda decoder descriptor
+ *****************************************************************************/
+struct decoder_sys_t
+{
+    bool                b_error;
+    vlc_object_t        *p_log;
+    VDADecoder          p_decoder ;
+
+    vlc_va_frame_t      *p_queue;
+    vlc_mutex_t         queueMutex;
+    int32_t             i_queueDepth;
+    
+    struct SwsContext   *p_sws_ctx;
+};
+
+/*****************************************************************************
+ * Local prototypes
+ *****************************************************************************/
+static int vda_CreateDecoder( decoder_t * , int , int , int , void *, int );
+static void vda_DestroyDecoder( decoder_t * );
+
+static CFDictionaryRef vda_DictionaryWithPTS( int64_t );
+static int64_t vda_PTSFromDictionary( CFDictionaryRef );
+
+static void vda_FrameQueuePop( decoder_t * );
+
+static int vda_GetSwsCpuMask( void );
+
+static void vda_CopyPicture( decoder_sys_t *,
+                             picture_t *,
+                             CVPixelBufferRef );
+
+typedef void (*VDADecoderOutputCallback)( void             *p_sys,
+                                          CFDictionaryRef  user_info,
+                                          OSStatus         status,
+                                          uint32_t         infoFlags,
+                                          CVImageBufferRef imageBuffer);
+
+static int  OpenDecoder ( vlc_object_t * );
+static void CloseDecoder ( vlc_object_t * );
+static picture_t *DecodeBlock ( decoder_t *, block_t ** );
+
+/*****************************************************************************
+ * Module descriptor
+ *****************************************************************************/
+vlc_module_begin ()
+set_category( CAT_INPUT )
+set_subcategory( SUBCAT_INPUT_VCODEC )
+set_description( N_("VDA video decoder") )
+set_capability( "decoder", 1000 )
+set_callbacks( OpenDecoder, CloseDecoder )
+add_shortcut( "vda" )
+vlc_module_end ()
+
+/****************************************************************************
+ * vda_DecoderOutputCallback:
+ ****************************************************************************/
+/****************************************************************************
+ * Allocate a new frame and populate it with some information
+ * this pointer to a vlc_va_frame_t type keeps track of the newest decompressed frame
+ * and is then inserted into a linked list of frame pointers depending on the display time
+ * stored in the user_info dictionary.
+ ****************************************************************************/
+static void vda_DecoderOutputCallback ( void             *p_sys,
+                                        CFDictionaryRef  user_info,
+                                        OSStatus         status,
+                                        uint32_t         infoFlags,
+                                        CVImageBufferRef imageBuffer)
+{
+    decoder_sys_t *p_decoder_sys = ( decoder_sys_t* ) p_sys;
+    
+    if ( NULL == imageBuffer ) 
+    {
+        /* FIXME: kVDADecodeInfo_FrameDropped not defined in VideoDecodeAcceleration/VDADecoder.h */
+        if ( 1UL << 1 & infoFlags )
+        {
+            msg_Dbg( p_decoder_sys->p_log, "frame dropped!" );
+        }
+        return;
+    }
+
+    if ( kCVPixelFormatType_422YpCbCr8 != CVPixelBufferGetPixelFormatType( imageBuffer ) )
+    {
+        msg_Err( p_decoder_sys->p_log, "image buffer format not '2vuy'." );
+        return;
+    }
+    
+    /* Create new frame. */
+    vlc_va_frame_t *newFrame = (vlc_va_frame_t *)calloc( sizeof( vlc_va_frame_t), 1 );
+    newFrame->p_next = NULL;
+    newFrame->buffer = CVPixelBufferRetain( imageBuffer );
+    newFrame->i_pts = vda_PTSFromDictionary( user_info );
+
+    /* since the frames we get may be in decode order rather than presentation order
+       our hypothetical callback places them in a queue of frames which will
+       hold them in display order. */
+       
+    vlc_mutex_lock( &p_decoder_sys->queueMutex );
+
+    vlc_va_frame_t *queueWalker = p_decoder_sys->p_queue;
+    
+    if ( !queueWalker || ( newFrame->i_pts < queueWalker->i_pts ) ) 
+    {
+        /* we have an empty queue, or this frame earlier than the current queue head */
+        newFrame->p_next = queueWalker;
+        p_decoder_sys->p_queue = newFrame;
+    } 
+    else
+    {
+        /* walk the queue and insert this frame where it belongs in display order */
+        bool frameInserted = false;
+        vlc_va_frame_t *nextFrame = NULL;
+
+        while ( !frameInserted ) 
+        {
+            nextFrame = queueWalker->p_next;
+            
+            if ( !nextFrame || ( newFrame->i_pts < nextFrame->i_pts ) ) 
+            {
+                /* if the next frame is the tail of the queue, or our new frame is ealier */
+                newFrame->p_next = nextFrame;
+                queueWalker->p_next = newFrame;
+                frameInserted = true;
+            }
+            queueWalker = nextFrame;
+        }
+    }
+
+    p_decoder_sys->i_queueDepth++;
+
+    vlc_mutex_unlock( &p_decoder_sys->queueMutex );
+}
+
+/*****************************************************************************
+ * vda_FrameQueuePop: pop the top frame from the queue and release it.
+ *****************************************************************************/
+static void vda_FrameQueuePop( decoder_t *p_dec )
+{
+    decoder_sys_t *p_sys = ( decoder_sys_t* ) p_dec->p_sys;
+    
+    if ( p_sys->p_queue == NULL ) 
+        return;
+    
+    vlc_mutex_lock( &p_sys->queueMutex );
+    
+    vlc_va_frame_t *top_frame = p_sys->p_queue;
+    p_sys->p_queue = (vlc_va_frame_t *)top_frame->p_next;
+    p_sys->i_queueDepth--;
+    
+    vlc_mutex_unlock( &p_sys->queueMutex );
+    
+    CVPixelBufferRelease( top_frame->buffer );
+    
+    free( top_frame );
+}
+
+/*****************************************************************************
+ * vda_DictionaryWithPTS: store the presentation time in the user info dictionary.
+ *****************************************************************************/
+static CFDictionaryRef vda_DictionaryWithPTS( int64_t i_pts )
+{
+    CFStringRef key = CFSTR( "VLC_VDA_DECODER_PTS_KEY" );
+    CFNumberRef value = CFNumberCreate( kCFAllocatorDefault, kCFNumberSInt64Type, &i_pts );
+    
+    return CFDictionaryCreate(kCFAllocatorDefault,
+                              (const void **)&key,
+                              (const void **)&value,
+                              1,
+                              &kCFTypeDictionaryKeyCallBacks,
+                              &kCFTypeDictionaryValueCallBacks);
+}
+
+/*****************************************************************************
+ * vda_PTSFromDictionary: get the presentation time from the user info dictionary.
+ *****************************************************************************/
+static int64_t vda_PTSFromDictionary( CFDictionaryRef user_info )
+{
+    CFNumberRef timeNumber = NULL;
+    int64_t outValue = 0;
+    
+    if ( NULL == user_info )
+        return 0;
+    
+    timeNumber = CFDictionaryGetValue( user_info, CFSTR( "VLC_VDA_DECODER_PTS_KEY" ) );
+    
+    if ( timeNumber )
+    {
+        CFNumberGetValue( timeNumber, kCFNumberSInt64Type, &outValue );
+    }
+    
+    return outValue;
+}
+
+/*****************************************************************************
+ * vda_CreateDecoder: create the video decoder acceleration.
+ *****************************************************************************/
+static int vda_CreateDecoder( decoder_t *p_dec, 
+                              int i_width,
+                              int i_height,
+                              int i_format,
+                              void *p_avc_data,
+                              int i_avc_data)
+{
+    OSStatus status = kVDADecoderNoErr;
+    CFNumberRef config_height = NULL;
+    CFNumberRef config_width = NULL;
+    CFNumberRef config_source_format = NULL;
+    CFDataRef config_avcC_data = NULL;
+    CFMutableDictionaryRef decoder_config_info = NULL;
+    
+    if ( i_format != 'avc1' )
+    {
+        msg_Err( p_dec, "source format is not h264." );
+        return VLC_EGENERIC;
+    }
+    
+    if ( p_avc_data == NULL )
+    {
+        msg_Err( p_dec, "vda decoder require avc data." );
+        return VLC_EGENERIC;
+    }
+    
+    decoder_config_info = ( CFDictionaryCreateMutable( kCFAllocatorDefault, 
+                                                       4,
+                                                       &kCFTypeDictionaryKeyCallBacks,
+                                                       &kCFTypeDictionaryValueCallBacks ) );
+    
+    config_height = CFNumberCreate( kCFAllocatorDefault, kCFNumberSInt32Type, &i_height );
+    config_width = CFNumberCreate( kCFAllocatorDefault, kCFNumberSInt32Type, &i_width );
+    config_source_format = CFNumberCreate( kCFAllocatorDefault, kCFNumberSInt32Type, &i_format );
+    config_avcC_data = CFDataCreate( kCFAllocatorDefault, p_avc_data, i_avc_data );
+    
+    CFDictionarySetValue( decoder_config_info, kVDADecoderConfiguration_Height, config_height );
+    CFDictionarySetValue( decoder_config_info, kVDADecoderConfiguration_Width, config_width );
+    CFDictionarySetValue( decoder_config_info, kVDADecoderConfiguration_SourceFormat, config_source_format );
+    CFDictionarySetValue( decoder_config_info, kVDADecoderConfiguration_avcCData, config_avcC_data );
+    
+    status = VDADecoderCreate( decoder_config_info,
+                               NULL, 
+                               (VDADecoderOutputCallback *)vda_DecoderOutputCallback,
+                               (void *)p_dec->p_sys,
+                               &p_dec->p_sys->p_decoder );
+    
+    if ( config_height ) CFRelease( config_height ) ;
+    if ( config_width ) CFRelease( config_width ) ;
+    if ( config_source_format ) CFRelease( config_source_format ) ;
+    if ( config_avcC_data ) CFRelease( config_avcC_data ) ;
+    if ( decoder_config_info ) CFRelease( decoder_config_info ) ;
+    
+    if ( kVDADecoderNoErr != status )
+    {
+        msg_Err( p_dec, "failed to create vda decoder: %d.", status );
+        return VLC_EGENERIC;
+    }
+    
+    return VLC_SUCCESS; 
+}
+
+/*****************************************************************************
+ * vda_DestroyDecoder: flush all queued frames and destroy the decoder.
+ *****************************************************************************/
+static void vda_DestroyDecoder( decoder_t *p_dec )
+{
+    decoder_sys_t *p_sys = (decoder_sys_t *)p_dec->p_sys;
+    OSStatus status = kVDADecoderNoErr;
+    
+    status = VDADecoderDestroy( p_sys->p_decoder );
+    
+    if ( kVDADecoderNoErr != status )
+    {
+        msg_Err( p_dec, "failed to destroy decoder: %d.", status );
+        return;
+    }
+    
+    msg_Dbg( p_dec, "decoder destroyed." );
+}
+
+/*****************************************************************************
+ * vda_GetSwsCpuMask: helper (taken from swscale.c)
+ *****************************************************************************/
+static int vda_GetSwsCpuMask(void)
+{
+    const unsigned int i_cpu = vlc_CPU();
+    int i_sws_cpu = 0;
+    
+    if( i_cpu & CPU_CAPABILITY_MMX )
+        i_sws_cpu |= SWS_CPU_CAPS_MMX;
+#if (LIBSWSCALE_VERSION_INT >= ((0<<16)+(5<<8)+0))
+    if( i_cpu & CPU_CAPABILITY_MMXEXT )
+        i_sws_cpu |= SWS_CPU_CAPS_MMX2;
+#endif
+    if( i_cpu & CPU_CAPABILITY_3DNOW )
+        i_sws_cpu |= SWS_CPU_CAPS_3DNOW;
+    
+    if( i_cpu & CPU_CAPABILITY_ALTIVEC )
+        i_sws_cpu |= SWS_CPU_CAPS_ALTIVEC;
+    
+    return i_sws_cpu;
+}
+
+/*****************************************************************************
+ * vda_CopyPicture: convert PIX_FMT_UYVY422 to PIX_FMT_YUV420P
+ *****************************************************************************/
+static void vda_CopyPicture( decoder_sys_t *p_sys,
+                             picture_t *p_pic,
+                             CVPixelBufferRef buffer )
+{
+    if ( p_sys->p_sws_ctx )
+    {
+        CVPixelBufferLockBaseAddress( buffer, 0 );
+        
+        uint8_t *p_src[]  = { CVPixelBufferGetBaseAddressOfPlane( buffer, 0 ), 0, 0, 0 };
+        int i_src_strides[] = { CVPixelBufferGetBytesPerRowOfPlane( buffer, 0 ), 0, 0, 0 };
+        
+        uint8_t *p_dst[]  = { p_pic->p[0].p_pixels, p_pic->p[1].p_pixels, p_pic->p[2].p_pixels, 0 };
+        int i_dst_strides[] = { p_pic->p[0].i_pitch, p_pic->p[1].i_pitch, p_pic->p[2].i_pitch, 0 };
+
+#if LIBSWSCALE_VERSION_INT  >= ((0<<16)+(5<<8)+0)
+        sws_scale(p_sys->p_sws_ctx, p_src, i_src_strides, 0, p_pic->format.i_height, p_dst, i_dst_strides);
+#else
+        sws_scale_ordered(p_sys->p_sws_ctx, p_src, i_src_strides, 0, p_pic->format.i_height, p_dst, i_dst_strides);
+#endif
+        CVPixelBufferUnlockBaseAddress( buffer, 0 );
+    }
+}
+
+/*****************************************************************************
+ * OpenDecoder: probe the decoder and return score
+ *****************************************************************************/
+static int OpenDecoder( vlc_object_t *p_this )
+{
+    decoder_t *p_dec = (decoder_t*)p_this;
+
+    if ( p_dec->fmt_in.i_codec != VLC_CODEC_H264 )
+    {
+        return VLC_EGENERIC;
+    }
+    
+    if ( p_dec->fmt_in.i_original_fourcc == VLC_FOURCC( 'a', 'v', 'c', '1') &&
+        p_dec->fmt_in.i_extra < 7 )
+    {
+        msg_Err( p_dec, "not valid avc." );
+        return VLC_EGENERIC;
+    }
+    
+    /* Allocate the memory needed to store the decoder's structure */
+    if ( ( p_dec->p_sys = malloc(sizeof(decoder_sys_t)) ) == NULL )
+        return VLC_ENOMEM;
+    
+    /* Create the va decoder. */
+    if ( vda_CreateDecoder(p_dec, 
+                           p_dec->fmt_in.video.i_width,
+                           p_dec->fmt_in.video.i_height,
+                           'avc1',
+                           p_dec->fmt_in.p_extra,
+                           p_dec->fmt_in.i_extra) )
+    {
+        free( p_dec->p_sys );
+        return VLC_EGENERIC;
+    }
+    
+    /* Set output properties */
+    es_format_Init( &p_dec->fmt_out, VIDEO_ES, VLC_CODEC_I420 );
+    p_dec->fmt_out.video.i_width = p_dec->fmt_in.video.i_width;
+    p_dec->fmt_out.video.i_height = p_dec->fmt_in.video.i_height;
+    p_dec->fmt_out.video.i_sar_num = p_dec->fmt_in.video.i_sar_num;
+    p_dec->fmt_out.video.i_sar_den = p_dec->fmt_in.video.i_sar_den;
+    
+    /* Set the decoder's structure properties */
+    p_dec->p_sys->p_log = p_this;
+    vlc_mutex_init( &p_dec->p_sys->queueMutex );
+    p_dec->p_sys->p_queue = NULL;
+    p_dec->p_sys->p_sws_ctx = sws_getContext( p_dec->fmt_in.video.i_width, p_dec->fmt_in.video.i_height, PIX_FMT_UYVY422,
+                                              p_dec->fmt_out.video.i_width, p_dec->fmt_out.video.i_height, PIX_FMT_YUV420P,
+                                              SWS_FAST_BILINEAR | vda_GetSwsCpuMask(),
+                                              NULL, NULL, NULL );
+
+    /* Set callbacks */
+    p_dec->pf_decode_video = DecodeBlock;
+    
+    msg_Dbg( p_dec, "vda decoder initialized." );
+
+    return VLC_SUCCESS;
+}
+
+/****************************************************************************
+ * DecodeBlock: the whole thing
+ ****************************************************************************
+ * This function must be fed with a complete compressed frame.
+ ****************************************************************************/
+static picture_t *DecodeBlock( decoder_t *p_dec, block_t **pp_block )
+{    
+    OSStatus status = kVDADecoderNoErr;
+    decoder_sys_t *p_sys = ( decoder_sys_t * ) p_dec->p_sys;
+    block_t *p_block;
+    
+    if ( !pp_block || !*pp_block )
+        return NULL;
+    
+    p_block = *pp_block;
+    p_sys->b_error = false;
+    
+    CFDictionaryRef block_pts = vda_DictionaryWithPTS(p_block->i_pts);
+    CFDataRef block_data = CFDataCreate(kCFAllocatorDefault, p_block->p_buffer, p_block->i_buffer);
+
+    status = VDADecoderDecode((VDADecoder)p_sys->p_decoder, 0, block_data, block_pts);
+    
+    if (block_pts) CFRelease(block_pts);
+    if (block_data) CFRelease(block_data);
+    
+    if (kVDADecoderNoErr != status)
+    {
+        msg_Err( p_dec, "decoder failed to decode : %d", status);
+    }
+    
+    if ( p_sys->i_queueDepth < 10 )
+    {
+        msg_Dbg( p_dec, "decoder is buffering...");
+        block_Release( p_block ); *pp_block = NULL;
+        return NULL;
+    }
+    
+    /* Get a new picture */
+    picture_t *p_pic = decoder_NewPicture( p_dec );
+
+    if ( !p_pic ) 
+    {
+        block_Release( p_block ); *pp_block = NULL;
+        return NULL;
+    }
+    
+    vlc_mutex_lock( &p_sys->queueMutex );
+
+    if ( !p_sys->p_queue )
+    {
+        vlc_mutex_unlock( &p_sys->queueMutex );
+        decoder_DeletePicture( p_dec, p_pic );
+        block_Release( p_block ); *pp_block = NULL;
+        return NULL;
+    }
+    
+    CVPixelBufferRef buffer = p_sys->p_queue->buffer;
+    p_pic->date = p_sys->p_queue->i_pts;
+ 
+    if ( !buffer ) 
+    {
+        decoder_DeletePicture( p_dec, p_pic );
+        vlc_mutex_unlock( &p_sys->queueMutex );
+        block_Release( p_block ); *pp_block = NULL;
+        return NULL;
+    }
+    
+    vlc_mutex_unlock( &p_sys->queueMutex );
+
+    vda_CopyPicture( p_sys, p_pic, buffer );
+
+    vda_FrameQueuePop( p_dec );
+    
+    block_Release( p_block ); *pp_block = NULL;
+
+    return p_pic;
+}
+
+/*****************************************************************************
+ * CloseDecoder: vda decoder destruction
+ *****************************************************************************/
+static void CloseDecoder( vlc_object_t *p_this )
+{
+    decoder_t *p_dec = ( decoder_t * ) p_this;
+    decoder_sys_t *p_sys = p_dec->p_sys;
+    
+    /* Destroy and release resources. */
+    vda_DestroyDecoder( p_dec ) ;   
+    
+    /* release vda structure */
+    while ( p_sys->p_queue != NULL )
+    {
+        vda_FrameQueuePop(p_dec);
+    }
+    
+    vlc_mutex_destroy( &p_sys->queueMutex );   
+    
+    if( p_sys->p_sws_ctx )
+        sws_freeContext( p_sys->p_sws_ctx );
+    
+    msg_Dbg( p_dec, "decoder is closing." );
+
+    free( p_sys );
+}
\ No newline at end of file
-- 
1.7.0.6

